/* Finding memory locations
 *
 * 15-122 Principles of Imperative Computations, Fall 2012
 */

#include <stdlib.h>
#include <stdio.h>
#include "contracts.h"

struct uba {
  size_t next;
  size_t limit;
  void** data;
};

typedef int(*int2int)(int x);

int inc(int x) { return x+1; }

int fib(int n) {
  REQUIRES(n >= 0);
  int A[n+2];            /* stack-allocated array (VLA) */
  A[0] = 0;
  A[1] = 1;
  for (int i = 0; i <= n-2; i++)
    A[i+2] = A[i] + A[i+1];
  return A[n];           /* deallocated just before actual return */
}

int main() {
  /* Everything has an address */
  /* We can point to anything */

  /* heap, arrays */
  int *A = calloc(10, sizeof(int));
  printf("A        = %p\n", (void*)A);
  printf("&A[1]    = %p\n", (void*)&A[1]);
  printf("A+1      = %p\n", (void*)(A+1));

  /* heap, struct */
  struct uba *U = malloc(sizeof(struct uba));
  U->next = 0;
  U->limit = 17;
  U->data = (void*)A;
  printf("U        = %p\n", (void*)U);
  printf("&U->data = %p\n", (void*)&U->data);
  printf("U->data  = %p\n", (void*)U->data);

  /* stack, variable */
  int x = 15122;
  int y = 15213;
  printf("&x       = %p\n", (void*)&x);
  printf("&y       = %p\n", (void*)&y);

  /* stack, array */
  int B[3] = {7, 92, 3};
  printf("B        = %p\n", B);
  printf("&B       = %p\n", (void*)&B);
  printf("&B[2]    = %p\n", (void*)&B[2]);
  /* and for the earlier, heap-allocated A */
  printf("&A       = %p\n", (void*)&A);

  /* data segment, read-only */
  char *s = "Are you sure?";
  printf("&s       = %p\n", (void*)&s);
  printf("s        = %p\n", (void*)s);

  /* text segment, read-only */
  printf("&inc     = %p\n", (void*)&inc);
  printf("inc[0]   = %x\n", *(int*)(void*)&inc);

  printf("fib(10)  = %d\n", fib(10));

  /* size of a function pointer */
  /* %zd prints a size_t value */
  printf("sizeof(int2int) = %zd\n", sizeof(int2int));
  return 0;
}
